package ru.dbotthepony.mc.otm.menu

import com.google.common.collect.ImmutableList
import net.minecraft.world.Container
import net.minecraft.world.SimpleContainer
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.Slot
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.kommons.util.Delegate
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.value
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.IMatteryUpgrade
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.UpgradeType
import ru.dbotthepony.mc.otm.capability.energy
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.container.EnhancedContainer
import ru.dbotthepony.mc.otm.container.IEnhancedContainer
import ru.dbotthepony.mc.otm.container.IFilteredContainerSlot
import ru.dbotthepony.mc.otm.container.ItemFilter
import ru.dbotthepony.mc.otm.container.ItemFilterSet
import ru.dbotthepony.mc.otm.container.UpgradeContainer
import ru.dbotthepony.mc.otm.container.util.containerSlotOrNull
import ru.dbotthepony.mc.otm.core.collect.ConditionalEnumSet
import ru.dbotthepony.mc.otm.core.immutableList
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.menu.input.BooleanInputWithFeedback
import ru.dbotthepony.mc.otm.menu.input.InstantBooleanInput
import ru.dbotthepony.mc.otm.menu.input.ItemFilterInput
import ru.dbotthepony.mc.otm.network.StreamCodecs
import ru.dbotthepony.mc.otm.player.IPlayerInventorySlot
import java.util.*
import java.util.function.BooleanSupplier
import java.util.function.DoubleSupplier
import java.util.function.IntSupplier
import java.util.function.Supplier
import kotlin.reflect.KMutableProperty0

/**
 * Make slots for single container
 */
inline fun <C : Container, S : Slot> makeSlots(container: C, initializer: (C, Int) -> S): ImmutableList<S> {
	return immutableList(container.containerSize) { initializer.invoke(container, it) }
}

/**
 * Make slots for list of containers with single slot in them
 */
inline fun <S : Slot> makeSlots(containers: List<Container>?, size: Int, initializer: (Container, Int) -> S): ImmutableList<S> {
	return immutableList(size) { initializer(containers?.get(it) ?: SimpleContainer(1), 0) }
}

open class MatteryMenuSlot(container: Container, index: Int, x: Int = 0, y: Int = 0) : Slot(container, index, x, y) {
	var ignoreSpectators = true

	open fun setupNetworkControls(menu: MatteryMenu) {
		val slot = containerSlotOrNull()

		if (slot is IFilteredContainerSlot && slot !is IPlayerInventorySlot) {
			menu.mSynchronizer.add(Delegate.Of(slot::filter), StreamCodecs.ITEM_FILTER)
		}
	}

	override fun setChanged() {
		if (container is IEnhancedContainer<*>) {
			(container as IEnhancedContainer<*>).setChanged(containerSlot)
		} else {
			super.setChanged()
		}
	}

	override fun mayPickup(player: Player): Boolean {
		return super.mayPickup(player) && (!ignoreSpectators || !player.isSpectator)
	}

	open fun canTakeItemForPickAll(): Boolean {
		return (container.containerSlotOrNull(slotIndex) as? IFilteredContainerSlot)?.filter?.allowAll != false
	}

	override fun getMaxStackSize(): Int {
		val container = container

		if (container is IEnhancedContainer<*>) {
			return container.getMaxStackSize(slotIndex, ItemStack.EMPTY)
		} else {
			return super.getMaxStackSize()
		}
	}

	override fun getMaxStackSize(itemStack: ItemStack): Int {
		val container = container

		if (container is IEnhancedContainer<*>) {
			return container.getMaxStackSize(slotIndex, itemStack)
		} else {
			return super.getMaxStackSize(itemStack)
		}
	}

	private fun isSameFilter(other: Slot): Boolean {
		val sSelf = containerSlotOrNull() as? IFilteredContainerSlot
		val sOther = other.containerSlotOrNull() as? IFilteredContainerSlot
		return sSelf?.filter == sOther?.filter
	}

	override fun isSameInventory(other: Slot): Boolean {
		return super.isSameInventory(other) && isSameFilter(other)
	}
}

open class UserFilteredMenuSlot(container: Container, index: Int, x: Int = 0, y: Int = 0) : MatteryMenuSlot(container, index, x, y) {
	var filterInput: MatteryMenu.PlayerInput<ItemFilter>? = null
		private set

	override fun setupNetworkControls(menu: MatteryMenu) {
		super.setupNetworkControls(menu)
		val slot = containerSlotOrNull()

		if (slot is IFilteredContainerSlot) {
			filterInput = menu.PlayerInput(StreamCodecs.ITEM_FILTER, handler = { slot.filter = it })
		}
	}
}

open class OutputMenuSlot(container: Container, index: Int, x: Int = 0, y: Int = 0, val onTake: (ItemStack) -> Unit = {}) : MatteryMenuSlot(container, index, x, y) {
	override fun mayPlace(itemStack: ItemStack): Boolean {
		return false
	}

	override fun onTake(pPlayer: Player, pStack: ItemStack) {
		super.onTake(pPlayer, pStack)
		this.onTake.invoke(pStack)
	}

	override fun onQuickCraft(pStack: ItemStack, pAmount: Int) {
		super.onQuickCraft(pStack, pAmount)
		this.onTake.invoke(pStack)
	}
}

open class BatteryMenuSlot(container: Container, index: Int, x: Int = 0, y: Int = 0, val direction: FlowDirection = FlowDirection.BI_DIRECTIONAL) : UserFilteredMenuSlot(container, index, x, y) {
	override fun mayPlace(itemStack: ItemStack): Boolean {
		if (!super.mayPlace(itemStack))
			return false

		val energy = itemStack.energy ?: return false
		if (energy is IMatteryEnergyStorage) return direction.test(energy.energyFlow)
		return direction.test(FlowDirection.of(energy.canReceive(), energy.canExtract()))
	}
}

open class ChemicalFuelMenuSlot(container: Container, index: Int, x: Int = 0, y: Int = 0) : MatteryMenuSlot(container, index, x, y) {
	override fun mayPlace(itemStack: ItemStack): Boolean {
		return super.mayPlace(itemStack) && itemStack.getBurnTime(null) > 0
	}
}

open class MatterContainerInputMenuSlot(
	container: Container,
	index: Int,
	x: Int = 0,
	y: Int = 0,
	val direction: FlowDirection = FlowDirection.BI_DIRECTIONAL
) : UserFilteredMenuSlot(container, index, x, y) {
	override fun mayPlace(itemStack: ItemStack): Boolean {
		val handler = itemStack.getCapability(MatteryCapability.MATTER_ITEM)
		return handler != null && super.mayPlace(itemStack) && this.direction.test(handler.matterFlow)
	}
}

open class PatternMenuSlot(container: Container, index: Int, x: Int = 0, y: Int = 0) : MatteryMenuSlot(container, index, x, y) {
	override fun mayPlace(itemStack: ItemStack): Boolean {
		return super.mayPlace(itemStack) && itemStack.getCapability(MatteryCapability.PATTERN_ITEM) != null
	}
}

open class DriveMenuSlot(container: Container, index: Int, x: Int = 0, y: Int = 0) : MatteryMenuSlot(container, index, x, y) {
	override fun mayPlace(itemStack: ItemStack): Boolean {
		return super.mayPlace(itemStack) && itemStack.getCapability(MatteryCapability.CONDENSATION_DRIVE) != null
	}
}

fun MatteryMenu.addFilterSlots(amount: Int, slots: Delegate<ItemFilterSet>?): ItemFilterInput {
	return ItemFilterInput(this, amount, slots)
}

fun MatteryMenu.addFilterSlots(amount: Int): ItemFilterInput {
	return addFilterSlots(amount, Delegate.Box(ItemFilterSet.EMPTY))
}

fun MatteryMenu.addFilterSlots(amount: Int, slots: KMutableProperty0<ItemFilterSet>?): ItemFilterInput {
	return addFilterSlots(amount, if (slots == null) null else Delegate.Of(slots))
}

val Slot.isOverCapacity: Boolean get() {
	return item.isNotEmpty && getMaxStackSize(item) <= item.count
}

/**
 * [openState] **is clientside only**, attempting to use it on server will result
 * in classloading exceptions.
 */
data class UpgradeSlots(
	val slots: List<MatteryMenuSlot>,
	val allowedTypes: Set<UpgradeType>,
	val openState: Delegate<Boolean>,
	val currentStats: IMatteryUpgrade,
	val areLocked: Supplier<Boolean>
)

fun MatteryMenu.makeUpgradeSlots(count: Int, container: UpgradeContainer?): UpgradeSlots {
	if (container != null) {
		require(count == container.containerSize) { "Upgrade container size ${container.containerSize} does not match with provided size $count" }
	}

	val shouldLockUpgradeSlots: Supplier<Boolean>
	val allowedTypes = EnumMap<UpgradeType, BooleanSupplier>(UpgradeType::class.java)

	if (container == null) {
		shouldLockUpgradeSlots = mSynchronizer.boolean()
	} else {
		shouldLockUpgradeSlots = mSynchronizer.computedBoolean(container.shouldLockUpgradeSlots)
	}

	for (value in UpgradeType.ALL) {
		val b = mSynchronizer.boolean().delegate.also {
			if (container != null) {
				it.accept(value in container.allowedUpgrades)
			}
		}

		allowedTypes[value] = BooleanSupplier { b.get() }
	}

	val syncContainer = container ?: EnhancedContainer.Simple(count)
	val isOpen = InstantBooleanInput(this)

	return UpgradeSlots(
		slots = immutableList(count) {
			object : MatteryMenuSlot(syncContainer, it) {
				init {
					mapQuickMoveToInventory(this)
				}

				override fun mayPlace(itemStack: ItemStack): Boolean {
					return super.mayPlace(itemStack) &&
						(itemStack.getCapability(MatteryCapability.UPGRADE)?.let { it.upgradeTypes.any { allowedTypes[it]!!.asBoolean } } == true) &&
						!shouldLockUpgradeSlots.get()
				}

				override fun mayPickup(player: Player): Boolean {
					return super.mayPickup(player) && !shouldLockUpgradeSlots.get()
				}
			}
		}.also { for (i in it.indices.reversed()) addStorageSlot(it[i], prepend = true, condition = isOpen) },

		areLocked = shouldLockUpgradeSlots,

		allowedTypes = ConditionalEnumSet(allowedTypes),
		openState = isOpen,
		currentStats = object : IMatteryUpgrade {
			override val upgradeTypes: Set<UpgradeType> = setOf()
			override val speedBonus: Double by mSynchronizer.computedDouble(DoubleSupplier { container?.speedBonus ?: 0.0 })
			override val processingItems: Int by mSynchronizer.computedInt(IntSupplier { container?.processingItems ?: 0 })
			override val energyStorageFlat: Decimal by mSynchronizer.computedDecimal { container?.energyStorageFlat ?: Decimal.ZERO }
			override val energyStorage: Decimal by mSynchronizer.computedDecimal { container?.energyStorage ?: Decimal.ZERO }
			override val energyConsumed: Decimal by mSynchronizer.computedDecimal { container?.energyConsumed ?: Decimal.ZERO }
			override val energyThroughputFlat: Decimal by mSynchronizer.computedDecimal { container?.energyThroughputFlat ?: Decimal.ZERO }
			override val energyThroughput: Decimal by mSynchronizer.computedDecimal { container?.energyThroughput ?: Decimal.ZERO }
			override val failureMultiplier: Double by mSynchronizer.computedDouble(DoubleSupplier { container?.failureMultiplier ?: 1.0 })
		}
	)
}
